
import collections
import operator
import time
import uuid

try:
    # Python 3.
    import queue
    from functools import reduce
except ImportError:
    # Python 2.
    import Queue as queue

import redis

from .client import default_client, transaction


####################################################################
#                                                                  #
#  The following functions are all used for creating methods that  #
#  get assigned to all the magic operator names for each type,     #
#  and make much more sense further down where they're applied.    #
#                                                                  #
####################################################################

def value_left(self, other):
    """
    Returns the value of the other type instance to use in an
    operator method, namely when the method's instance is on the
    left side of the expression.
    """
    return other.value if isinstance(other, self.__class__) else other


def value_right(self, other):
    """
    Returns the value of the type instance calling an to use in an
    operator method, namely when the method's instance is on the
    right side of the expression.
    """
    return self if isinstance(other, self.__class__) else self.value


def op_left(op):
    """
    Returns a type instance method for the given operator, applied
    when the instance appears on the left side of the expression.
    """
    def method(self, other):
        return op(self.value, value_left(self, other))
    return method


def op_right(op):
    """
    Returns a type instance method for the given operator, applied
    when the instance appears on the right side of the expression.
    """
    def method(self, other):
        return op(value_left(self, other), value_right(self, other))
    return method


def inplace(method_name):
    """
    Returns a type instance method that will call the given method
    name, used for inplace operators such as __iadd__ and __imul__.
    """
    def method(self, other):
        getattr(self, method_name)(value_left(self, other))
        return self
    return method


#####################################################################
#                                                                   #
#  Base class / groupings of logical operators that types inherit.  #
#                                                                   #
#####################################################################

class Base(object):
    """
    Base type that all others inherit. Contains the basic comparison
    operators as well as the dispatch for proxying to methods on the
    Redis client.
    """

    def __init__(self, initial=None, key=None, client=None):
        self.client = client  # Must be first.
        self.key = key or str(uuid.uuid4())
        if initial is not None:
            if key is None:
                self.value = initial
            else:
                # Ensure previous value removed if key and initial
                # value provided.
                with transaction():
                    self.delete()
                    self.value = initial

    __eq__ = op_left(operator.eq)
    __lt__ = op_left(operator.lt)
    __le__ = op_left(operator.le)
    __gt__ = op_left(operator.gt)
    __ge__ = op_left(operator.ge)

    def __repr__(self):
        bits = (self.__class__.__name__, repr(self.value), self.key)
        return "%s(%s, '%s')" % bits

    def __getattr__(self, name):
        return self._dispatch(name)

    def _dispatch(self, name):
        try:
            func = getattr(self.client or default_client(), name)
        except AttributeError:
            raise
        return lambda *a, **k: func(self.key, *a, **k)


class Bitwise(Base):
    """
    Base class for bitwise types and relevant operators.
    """
    __and__       = op_left(operator.and_)
    __or__        = op_left(operator.or_)
    __xor__       = op_left(operator.xor)
    __lshift__    = op_left(operator.lshift)
    __rshift__    = op_left(operator.rshift)
    __rand__      = op_right(operator.and_)
    __ror__       = op_right(operator.or_)
    __rxor__      = op_right(operator.xor)
    __rlshift__   = op_right(operator.lshift)
    __rrshift__   = op_right(operator.rshift)


class Sequential(Base):
    """
    Base class for sequence types and relevant operators.
    """
    __add__       = op_left(operator.add)
    __mul__       = op_left(operator.mul)
    __radd__      = op_right(operator.add)
    __rmul__      = op_right(operator.mul)


class Numeric(Base):
    """
    Base class for numeric types and relevant operators.
    """
    __add__       = op_left(operator.add)
    __sub__       = op_left(operator.sub)
    __mul__       = op_left(operator.mul)
    __floordiv__  = op_left(operator.floordiv)
    __truediv__   = op_left(operator.truediv)
    __mod__       = op_left(operator.mod)
    __divmod__    = op_left(divmod)
    __pow__       = op_left(operator.pow)
    __radd__      = op_right(operator.add)
    __rsub__      = op_right(operator.sub)
    __rmul__      = op_right(operator.mul)
    __rtruediv__  = op_right(operator.truediv)
    __rfloordiv__ = op_right(operator.floordiv)
    __rmod__      = op_right(operator.mod)
    __rdivmod__   = op_right(divmod)
    __rpow__      = op_right(operator.pow)
    __iadd__      = inplace("incr")
    __isub__      = inplace("decr")
    __imul__      = inplace("number_multiply")
    __idiv__      = inplace("number_divide")
    __ifloordiv__ = inplace("number_floordiv")
    __imod__      = inplace("number_mod")
    __ipow__      = inplace("number_pow")


try:
    # Python 2.
    Numeric.__div__  = op_left(operator.div)
    Numeric.__rdiv__ = op_right(operator.div)
except AttributeError:
    pass


####################################################
#                                                  #
#  Python types that map directly to Redis types.  #
#                                                  #
####################################################

class List(Sequential):
    """
    Redis list <-> Python list
    """

    @property
    def value(self):
        return self[:]

    @value.setter
    def value(self, value):
        self.extend(value)

    __iadd__ = inplace("extend")
    __imul__ = inplace("list_multiply")

    def __len__(self):
        return self.llen()

    def __setitem__(self, i, item):
        try:
            self.lset(i, item)
        except redis.exceptions.ResponseError:
            raise IndexError

    def __getitem__(self, i):
        if isinstance(i, slice):
            start = i.start if i.start is not None else 0
            stop = i.stop if i.stop is not None else 0
            return self.lrange(start, stop - 1)
        item = self.lindex(i)
        if item is None:
            raise IndexError
        return item

    def __delitem__(self, i):
        self.pop(i)

    def __iter__(self):
        return iter(self.value)

    def append(self, item):
        self.extend([item])

    def extend(self, other):
        self.rpush(*other)

    def insert(self, i, item):
        if i == 0:
            self.lpush(item)
        else:
            self.list_insert(i, item)

    def pop(self, i=-1):
        if i == -1:
            return self.rpop()
        elif i == 0:
            return self.lpop()
        else:
            return self.list_pop(i)

    def reverse(self):
        self.list_reverse()

    def index(self, item):
        return self.value.index(item)

    def count(self, item):
        return self.value.count(item)

    def sort(self, reverse=False):
        self._dispatch("sort")(desc=reverse, store=self.key, alpha=True)


class Set(Bitwise):
    """
    Redis set <-> Python set
    """

    @property
    def value(self):
        return self.smembers()

    @value.setter
    def value(self, item):
        self.update(item)

    def _all_redis(self, sets):
        return all([isinstance(s, self.__class__) for s in sets])

    def _to_keys(self, sets):
        return [s.key for s in sets]

    __iand__ = inplace("intersection_update")
    __ior__  = inplace("update")
    __ixor__ = inplace("symmetric_difference_update")
    __isub__ = inplace("difference_update")
    __rsub__ = op_right(operator.sub)

    def __and__(self, value):
        return self.intersection(value)

    def __or__(self, value):
        return self.union(value)

    def __xor__(self, value):
        return self.symmetric_difference(value)

    def __sub__(self, value):
        return self.difference(value)

    def __len__(self):
        return self.scard()

    def __contains__(self, item):
        return self.sismember(item)

    def __iter__(self):
        return iter(self.value)

    def add(self, item):
        self.update([item])

    def update(self, *sets):
        self.sadd(*reduce(operator.or_, sets))

    def pop(self):
        return self.spop()

    def clear(self):
        self.delete()

    def remove(self, item):
        if self.srem(item) == 0:
            raise KeyError(item)

    def discard(self, item):
        try:
            self.remove(item)
        except KeyError:
            pass

    def intersection(self, *sets):
        if self._all_redis(sets):
            return self.sinter(*self._to_keys(sets))
        else:
            return reduce(operator.and_, (self.value,) + sets)

    def intersection_update(self, *sets):
        if self._all_redis(sets):
            self.sinterstore(self.key, *self._to_keys(sets))
        else:
            sets = list(reduce(operator.and_, sets))
            self.set_intersection_update(*sets)
        return self

    def union(self, *sets):
        if self._all_redis(sets):
            return self.sunion(*self._to_keys(sets))
        else:
            return reduce(operator.or_, (self.value,) + sets)

    def difference(self, *sets):
        if self._all_redis(sets):
            return self.sdiff(*self._to_keys(sets))
        else:
            return reduce(operator.sub, (self.value,) + sets)

    def difference_update(self, *sets):
        if self._all_redis(sets):
            self.sdiffstore(self.key, *self._to_keys(sets))
        else:
            key = str(uuid.uuid4())
            flattened = [key]
            for s in sets:
                flattened.extend(s)
                flattened.append(key)
            self.set_difference_update(*flattened)
        return self

    def symmetric_difference(self, other):
        if isinstance(other, self.__class__):
            return set(self.set_symmetric_difference("return", other.key))
        else:
            return self.value ^ other

    def symmetric_difference_update(self, other):
        if isinstance(other, self.__class__):
            self.set_symmetric_difference("update", other.key)
        else:
            self.set_symmetric_difference("create", *other)
        return self

    def isdisjoint(self, other):
        return not self.intersection(other)

    def issubset(self, other):
        return self <= other

    def issuperset(self, other):
        return self >= other


class Dict(Base):
    """
    Redis hash <-> Python dict
    """

    @property
    def value(self):
        return self.hgetall()

    @value.setter
    def value(self, value):
        if not isinstance(value, dict):
            try:
                value = dict(value)
            except TypeError:
                value = None
        if value:
            self.update(value)

    def __len__(self):
        return self.hlen()

    def __contains__(self, key):
        return self.hexists(key)

    def __iter__(self):
        return self.iterkeys()

    def __setitem__(self, key, value):
        self.hset(key, value)

    def __getitem__(self, key):
        value = self.get(key)
        if value is None:
            raise KeyError(key)
        return value

    def __delitem__(self, key):
        if self.hdel(key) == 0:
            raise KeyError(key)

    def update(self, value):
        self.hmset(value)

    def keys(self):
        return self.hkeys()

    def values(self):
        return self.hvals()

    def items(self):
        return self.value.items()

    def iterkeys(self):
        return iter(self.keys())

    def itervalues(self):
        return iter(self.values())

    def iteritems(self):
        return iter(self.items())

    def setdefault(self, key, value=None):
        if self.hsetnx(key, value) == 1:
            return value
        else:
            return self.get(key)

    def get(self, key, default=None):
        value = self.hget(key)
        return value if value is not None else default

    def has_key(self, key):
        return key in self

    def copy(self):
        return self.__class__(self.value)

    def clear(self):
        self.delete()

    @classmethod
    def fromkeys(cls, *args):
        if len(args) == 1:
            args += ("",)
        return cls({}.fromkeys(*args))


class String(Sequential):
    """
    Redis string <-> Python string (although mutable).
    """

    @property
    def value(self):
        return self.get() or ""

    @value.setter
    def value(self, value):
        if value:
            self.set(value)

    __iadd__ = inplace("append")
    __imul__ = inplace("string_multiply")

    def __len__(self):
        return self.strlen()

    def __setitem__(self, i, s):
        if isinstance(i, slice):
            start = i.start if i.start is not None else 0
            stop = i.stop
        else:
            start = i
            stop = None
        if stop is not None and stop < start + len(s):
            self.string_setitem(start, stop, s)
        else:
            self.setrange(start, s)

    def __getitem__(self, i):
        if not isinstance(i, slice):
            i = slice(i, i + 1)
        start = i.start if i.start is not None else 0
        stop = i.stop if i.stop is not None else 0
        s = self.getrange(start, stop - 1)
        if not s:
            raise IndexError
        return s

    def __iter__(self):
        return iter(self.value)


class ImmutableString(String):
    """
    Redis string <-> Python string (actually immutable).
    """

    def __iadd__(self, other):
        self.key = self.__class__(self + other).key
        return self

    def __imul__(self, i):
        self.key = self.__class__(self * i).key
        return self

    def __setitem__(self, i):
        raise TypeError


class Int(Numeric, Bitwise):
    """
    Redis integer <-> Python integer.
    """

    @property
    def value(self):
        return int(float(self.get() or 0))

    @value.setter
    def value(self, value):
        if value is not None:
            self.set(value)

    __iand__    = inplace("number_and")
    __ior__     = inplace("number_or")
    __ixor__    = inplace("number_xor")
    __ilshift__ = inplace("number_lshift")
    __irshift__ = inplace("number_rshift")


class Float(Numeric):
    """
    Redis float <-> Python float.
    """

    @property
    def value(self):
        return float(self.get() or 0)

    @value.setter
    def value(self, value):
        if value is not None:
            self.set(value)

    __iadd__ = inplace("incrbyfloat")

    def __isub__(self, f):
        self.incrbyfloat(f * -1)
        return self


###################################################################
#                                                                 #
#  Following are the types found in the Python standard library,  #
#  that can be represented by extending / composing the above     #
#  types. First up: Queue module.                                 #
#                                                                 #
###################################################################

class Queue(List):
    """
    Redis list <-> Python list <-> Python's Queue.
    """

    maxsize = 0

    def __init__(self, maxsize=None, **kwargs):
        if maxsize is not None:
            self.maxsize = maxsize
        super(Queue, self).__init__(**kwargs)

    @property
    def queue(self):
        return self

    def qsize(self):
        return len(self)

    def empty(self):
        return self.qsize() == 0

    def full(self):
        return self.maxsize > 0 and self.qsize() >= self.maxsize

    def put(self, item, block=True, timeout=None):
        if self.maxsize == 0:
            self.append(item)
        else:
            if not block:
                timeout = 0
            start = time.time()
            while True:
                if self.queue_put(item, self.maxsize):
                    break
                if timeout is not None and time.time() - start >= timeout:
                    raise queue.Full
                time.sleep(.1)

    def put_nowait(self, item):
        self.put(item, block=False)

    def get(self, block=True, timeout=None):
        if block:
            item = self.blpop(timeout=timeout)
            if item is not None:
                item = item[1]
        else:
            item = self.pop()
        if item is None:
            raise queue.Empty
        return item

    def get_nowait(self):
        return self.get(block=False)

    def join(self):
        while not self.empty():
            sleep(.1)


class LifoQueue(Queue):
    """
    Redis list <-> Python list <-> Python's Queue.LifoQueue.
    """

    def append(self, item):
        self.lpush(item)


class SetQueue(Queue):
    """
    Redis list + Redis set <-> Queue with only unique items.
    """

    def __init__(self, *args, **kwargs):
        super(SetQueue, self).__init__(*args, **kwargs)
        self.set = Set(key="%s-set" % self.key)

    def get(self, *args, **kwargs):
        item = super(SetQueue, self).get(*args, **kwargs)
        self.set.remove(item)
        return item

    def put(self, item, *args, **kwargs):
        if self.set.sadd(item) > 0:
            super(SetQueue, self).put(item, *args, **kwargs)

    def delete(self):
        self._dispatch("delete")
        self.set.delete()


class LifoSetQueue(LifoQueue, SetQueue):
    """
    Redis list + Redis set <-> LifoQueue with only unique items.
    """
    pass


####################################################################
#                                                                  #
#  Next up, some lock structures from the threading module. These  #
#  are all backed by the above Queue class, since it provides the  #
#  blocking / non-blocking mechanics desired.                      #
#                                                                  #
####################################################################

class BoundedSemaphore(Queue):
    """
    Redis list <-> Python list <-> Queue <-> threading.BoundedSemaphore.

    BoundedSemaphore's ``value`` arg maps to Queue's ``maxsize``.
    BoundedSemaphore's acquire/release methods maps to Queue's put/get
    methods repectively, providing blocking/timeout mechanics.
    """

    maxsize = 1

    def __init__(self, value=None, **kwargs):
        super(BoundedSemaphore, self).__init__(value, **kwargs)

    def acquire(self, block=True, timeout=None):
        try:
            self.put(1, block, timeout)
        except queue.Full:
            return False
        return True

    def release(self):
        try:
            self.get(block=False)
        except queue.Empty:
            raise RuntimeError("Cannot release unacquired lock")

    def __enter__(self):
        self.acquire()

    def __exit__(self, t, v, tb):
        self.release()


class Semaphore(BoundedSemaphore):
    """
    Redis list <-> Python list <-> Queue <-> threading.Semaphore.

    Same implementation as BoundedSemaphore, but without a queue size.
    """

    def release(self):
        try:
            super(Semaphore, self).release()
        except RuntimeError:
            pass


class Lock(BoundedSemaphore):
    """
    Redis list <-> Python list <-> Queue <-> threading.Lock.

    Same implementation as BoundedSemaphore, but with a fixed
    queue size of 1.
    """

    def __init__(self, **kwargs):
        kwargs["value"] = None
        super(Lock, self).__init__(**kwargs)


class RLock(Lock):
    """
    Redis list <-> Python list <-> Queue <-> threading.RLock.

    Same implementation as BoundedSemaphore, but with a fixed
    queue size of 1, and as per re-entrant locks, can be acquired
    multiple times.
    """

    def __init__(self, *args, **kwargs):
        self.acquires = 0
        super(RLock, self).__init__(*args, **kwargs)

    def acquire(self, *args, **kwargs):
        result = True
        if self.acquires == 0:
            result = super(RLock, self).acquire(*args, **kwargs)
        if result:
            self.acquires += 1
        return result

    def release(self):
        if self.acquires > 1:
            self.acquires -= 1
            if self.acquires > 0:
                return
        super(RLock, self).release()


#####################################################################
#                                                                   #
#  Some members from Python's collections standard library module.  #
#                                                                   #
#####################################################################


class DefaultDict(Dict):
    """
    Redis hash <-> Python dict <-> Python's collections.DefaultDict.
    """

    def __init__(self, default_factory, *args, **kwargs):
        self.default_factory = default_factory
        super(DefaultDict, self).__init__(*args, **kwargs)

    def __getitem__(self, key):
        return self.setdefault(key, self.default_factory())


class MultiSet(Dict):
    """
    Redis hash <-> Python dict <-> Python's collections.Counter.
    """

    def __init__(self, iterable=None, key=None, **kwargs):
        super(MultiSet, self).__init__(key=key)
        self.update(iterable=iterable, **kwargs)

    @property
    def value(self):
        value = super(MultiSet, self).value
        kwargs = dict([(k, int(v)) for k, v in value.items()])
        return collections.Counter(**kwargs)

    __add__  = op_left(operator.add)
    __sub__  = op_left(operator.sub)
    __and__  = op_left(operator.and_)
    __or__   = op_left(operator.or_)
    __radd__ = op_right(operator.add)
    __rsub__ = op_right(operator.sub)
    __rand__ = op_right(operator.and_)
    __ror__  = op_right(operator.or_)
    __iadd__ = inplace("update")
    __isub__ = inplace("subtract")
    __iand__ = inplace("intersection_update")
    __ior__  = inplace("union_update")

    # Return 0 as a default, which allows bitwise ops to work correctly
    # in Python 3, as its Counter type no longer supports working with
    # missing values.
    def __getitem__(self, name):
        try:
            return super(MultiSet, self).__getitem__(name)
        except KeyError:
            return 0

    def __delitem__(self, name):
        try:
            super(MultiSet, self).__delitem__(name)
        except KeyError:
            pass

    def __repr__(self):
        bits = (self.__class__.__name__, repr(dict(self.value)), self.key)
        return "%s(%s, '%s')" % bits

    def values(self):
        values = super(MultiSet, self).values()
        return [int(v) for v in values]

    def get(self, key, default=None):
        value = self.hget(key)
        return int(value) if value is not None else default

    def _merge(self, iterable=None, **kwargs):
        if iterable:
            try:
                try:
                    # Python 2.
                    items = iterable.iteritems()
                except AttributeError:
                    # Python 3.
                    items = iterable.items()
            except AttributeError:
                for k in iterable:
                    kwargs[k] = kwargs.get(k, 0) + 1
            else:
                for k, v in items:
                    kwargs[k] = kwargs.get(k, 0) + v
        return kwargs.items()

    def _flatten(self, iterable, **kwargs):
        for k, v in self._merge(iterable, **kwargs):
            yield k
            yield v

    def _update(self, iterable, multiplier, **kwargs):
        for k, v in self._merge(iterable, **kwargs):
            self.hincrby(k, v * multiplier)

    def update(self, iterable=None, **kwargs):
        self._update(iterable, 1, **kwargs)

    def subtract(self, iterable=None, **kwargs):
        self._update(iterable, -1, **kwargs)

    def intersection_update(self, iterable=None, **kwargs):
        self.multiset_intersection_update(*self._flatten(iterable, **kwargs))

    def union_update(self, iterable=None, **kwargs):
        self.multiset_union_update(*self._flatten(iterable, **kwargs))

    def elements(self):
        for k, count in self.iteritems():
            for i in range(count):
                yield k

    def most_common(self, n=None):
        values = sorted(self.iteritems(), key=lambda v: v[1], reverse=True)
        if n:
            values = values[:n]
        return values

collections.abc.MutableMapping.register(MultiSet)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import json
from redis import Redis
from typing import (
        List,
        Optional,
        Tuple,
        )


class MultiWaitTask:
    """
    wait multi keys for handler
    """

    def __init__(self, client: Redis, keys: List[str], batch_size: int = 10):
        self.client = client
        self.keys = keys
        self.batch_size = batch_size

    def get_tasks(self) -> Optional[Tuple[str, List[dict]]]:
        result = self.client.blmpop(
                10,
                len(self.keys),
                *self.keys,
                direction="LEFT",
                count=self.batch_size,
        )
        if not result:
            return None
        return result[0], [
                json.loads(i)
                for i in result[1]
        ]

    def add_task(self, key, info: dict):
        self.client.rpush(key, json.dumps(info, ensure_ascii=False, indent=4))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Xiang Wang <ramwin@qq.com>


import datetime
import time

from typing import List
from redis import Redis


class DebounceTask:
    DELTA = int(datetime.datetime(2025, 1, 1, 0, 0, 0).timestamp())

    def __init__(self, client: Redis, key: str, timeout: int):
        """
        timeout: 1means 0.1s
        """
        self.client = client
        self.key = key
        self.timeout = timeout

    def get_time(self) -> int:
        return int((time.time() - self.DELTA) * 10)

    def add_task(self, taskid: str):
        self.client.zadd(self.key, {taskid: self.get_time()}, nx=True)

    def pop_tasks(self, max_wait: int=0) -> List[str]:
        """
        params:
            max_wait: 1 means 0.1 seconds
        """
        tasks = self._pop_tasks()
        if tasks or (not max_wait):
            return tasks
        first_task = self.client.zrange(self.key, 0, 0, withscores=True)
        if not first_task:
            time.sleep(max_wait * 0.1)
            return []
        add_time = first_task[0][1]
        now = self.get_time()
        need_wait = max(0, self.timeout - (now - add_time))
        if need_wait <= max_wait:
            time.sleep(need_wait / 10)
            return self._pop_tasks()
        return self._pop_tasks()

    def _pop_tasks(self) -> List[str]:
        now = self.get_time()
        delete_before = now - self.timeout

        with self.client.lock(self.key + ":lock", timeout=5):
            taskids, _ = self.client.pipeline()\
                .zrangebyscore(self.key, min=0, max=delete_before)\
                .zremrangebyscore(self.key, min=0, max=delete_before)\
                .execute()
            return taskids



#!/usr/bin/env python3
# -*- coding: utf-8 -*-



from redis import Redis


class UniqueId:
    """
    When you have many databases, you may want to create an same uniqueid for a unique id.

        e.g.
            there were three databases named: students_info, students_activity, students_history, both shared the same table: students
            when you create a new student, you want to retrieve the student's id.

            unique_id = UniqueId(client=Redis(decode_respones=True), klass="students")
            unique_id.get_or_create("student's id card No.1") => 3 unique
    """

    def __init__(self, client: Redis, klass: str):
        self.client = client
        self.klass = klass
        self.value_key = klass + ":auto"

    def get_or_create(self, key: str) -> int:
        redis_key = self._get_redis_key(key)
        if (created := self.client.get(redis_key)):
            return int(created)
        newid = self.client.incr(self.value_key)
        valid = self.client.set(redis_key, newid, nx=True)
        if valid:
            return newid
        return int(self.client.get(redis_key))

    def set(self, key: str, id: int):
        """override the int for a key"""
        redis_key = self._get_redis_key(key)
        self.client.set(redis_key, id)

    def incr_to(self, id: int):
        self.client.set(self.value_key, id)

    def _get_redis_key(self, key: str):
        return self.klass + ":id:" + key




import contextlib
import os
import threading

import redis


class HotClient(redis.Redis):
    """
    A Redis client wrapper that loads Lua functions and creates
    client methods for calling them.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("decode_responses", True)
        super(HotClient, self).__init__(*args, **kwargs)
        requires_luabit = ("number_and", "number_or", "number_xor",
                           "number_lshift", "number_rshift")
        with open(self._get_lua_path("bit.lua")) as f:
            luabit = f.read()
        for name, snippet in self._get_lua_funcs():
            if name in requires_luabit:
                snippet = luabit + snippet
            self._create_lua_method(name, snippet)

    def _get_lua_path(self, name):
        """
        Joins the given name with the relative path of the module.
        """
        parts = (os.path.dirname(os.path.abspath(__file__)), "lua", name)
        return os.path.join(*parts)

    def _get_lua_funcs(self):
        """
        Returns the name / code snippet pair for each Lua function
        in the atoms.lua file.
        """
        with open(self._get_lua_path("atoms.lua")) as f:
            for func in f.read().strip().split("function "):
                if func:
                    bits = func.split("\n", 1)
                    name = bits[0].split("(")[0].strip()
                    snippet = bits[1].rsplit("end", 1)[0].strip()
                    yield name, snippet

    def _create_lua_method(self, name, code):
        """
        Registers the code snippet as a Lua script, and binds the
        script to the client as a method that can be called with
        the same signature as regular client methods, eg with a
        single key arg.
        """
        script = self.register_script(code)
        setattr(script, "name", name)  # Helps debugging redis lib.
        method = lambda key, *a, **k: script(keys=[key], args=a, **k)
        setattr(self, name, method)


_thread = threading.local()
_config = {}


def default_client():
    try:
        _thread.client
    except AttributeError:
        setattr(_thread, "client", HotClient(**_config))
    return _thread.client


def configure(**config):
    global _config
    _config = config


@contextlib.contextmanager
def transaction():
    """
    Swaps out the current client with a pipeline instance,
    so that each Redis method call inside the context will be
    pipelined. Once the context is exited, we execute the pipeline.
    """
    client = default_client()
    _thread.client = client.pipeline()
    try:
        yield
        _thread.client.execute()
    finally:
        _thread.client = client





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Xiang Wang <ramwin@qq.com>


from redis import Redis


class RedisRange:
    """
    In many scenario, you want to handle a task from start to end;
    e.g.
        save all database value to cache from id=1 to largest

        updated_range = RedisRange(redis_client=Reids(decode_responses=True), key="student_name_update_range")
        student_names = DelayButFastSet(redis_client=Reids(decode_responses=True), key="student_name_set")

        when user created:
            student_names.add(user.name)
            for old_user_id in range(updated_range.get_max() + 1, user.id - 1):
                student_names.add(User(id=old_user_id).name)
            updated_range.upper_to(user.id)

            updated_min = updated_range.get_min()
            to_update_min = updated_min - 10
            for old_user_id in range(to_update_min, updated_min):
                student_names.add(User(id=old_user_id).name)
            updated_range.lower_to(to_update_min)
    """

    def __init__(self, redis_client: Redis, key: str):
        assert redis_client.get_encoder().decode_responses is True
        self.redis_client = redis_client
        self.min_key = f"{key}:min"  # min value include
        self.max_key = f"{key}:max"  # max value include
        self.lock_key = f"{key}:lock"
        self.pending_key = f"{key}:pending"
        self.inited = self.redis_client.exists(self.max_key)

    def get_min_max(self) -> (int, int):
        assert self.inited
        a, b = self.redis_client.pipeline()\
                .get(self.min_key)\
                .get(self.max_key)\
                .execute()
        return int(a), int(b)

    def get_min(self) -> int:
        assert self.inited
        return int(self.redis_client.get(self.min_key))

    def get_max(self) -> int:
        assert self.inited
        return int(self.redis_client.get(self.max_key))

    def upper_to(self, stop: int) -> None:
        """
        increate max to stop 
        the value in (max: stop) will be added into pending
        """
        if not self.inited:
            self.init_with(stop)
        current_min, current_max = self.get_min_max()
        if stop < current_min:
            raise ValueError
        if stop <= current_max:
            return
        if stop == current_max + 1:
            self.redis_client.set(self.max_key, stop)
            return
        pending_ids = range(current_max+1, stop)
        self.redis_client.pipeline()\
            .set(self.max_key, stop)\
            .sadd(self.pending_key, *pending_ids)\
            .execute()

    def lower_to(self, start: int) -> None:
        """
        decrease min to start
        the value in [start: min) will be added into pending
        """
        if not self.inited:
            self.init_with(start)
        current_min, current_max = self.get_min_max()
        if start > current_max:
            raise ValueError
        elif start >= current_min:
            return
        pending_ids = range(start, current_min)
        self.redis_client.pipeline()\
                .set(self.min_key, start)\
                .sadd(self.pending_key, *pending_ids)\
                .execute()

    def init_with(self, value: int):
        with self.redis_client.lock(self.lock_key):
            if self.redis_client.exists(self.max_key):
                self.inited = True
                return
            self.redis_client.pipeline()\
                    .set(self.max_key, value)\
                    .set(self.min_key, value)\
                    .execute()
            self.inited = True

    def reset(self):
        self.redis_client.pipeline()\
                .delete(self.min_key)\
                .delete(self.max_key)\
                .delete(self.lock_key)\
                .delete(self.pending_key)\
                .execute()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import logging
import unittest
from redis import Redis

from hot_redis.multi_wait_task import MultiWaitTask


LOGGER = logging.getLogger(__name__)


class Test(unittest.TestCase):

    def test1(self):
        task = MultiWaitTask(
                client=Redis(decode_responses=True),
                keys=["list1", "list2"], batch_size=3
        )
        task.add_task("list1", ["list1task_a"])
        task.add_task("list1", ["list1task_b"])
        task.add_task("list2", ["list2task"])
        LOGGER.error(task.get_tasks())
        LOGGER.error(task.get_tasks())
        LOGGER.error(task.get_tasks())



#!/usr/bin/env python

import collections
import os
import time
import unittest
import hot_redis


# Env var specifying we're dealing with a server that doesn't support
# Lua, like an old version of Redis, or some kind of Redis clone.
TEST_NO_LUA = os.environ.get("HOT_REDIS_TEST_NO_LUA") == "1"

# Env var specifying the precision for assertAlmostEqual - we may want
# to define this for alternate Lua implementations, like LuaJ.
TEST_PRECISION = int(os.environ.get("HOT_REDIS_TEST_PRECISION", 0)) or None

keys = []

def base_wrapper(init):
    def wrapper(*args, **kwargs):
        init(*args, **kwargs)
        keys.append(args[0].key)
    return wrapper

hot_redis.Base.__init__ = base_wrapper(hot_redis.Base.__init__)
if TEST_NO_LUA:
    hot_redis.HotClient._create_lua_method = lambda *args, **kwargs: None


class BaseTestCase(unittest.TestCase):

    def tearDown(self):
        client = hot_redis.default_client()
        while keys:
            client.delete(keys.pop())

    # Removed in Python 3.
    def assertItemsEqual(self, a, b):
        self.assertEqual(sorted(a), sorted(b))

    # Configurable precision.
    def assertAlmostEqual(self, a, b):
        kwargs = {"places": TEST_PRECISION}
        return super(BaseTestCase, self).assertAlmostEqual(a, b, **kwargs)


class ListTests(BaseTestCase):

    def test_initial(self):
        a = ["wagwaan", "hot", "skull"]
        b = hot_redis.List(a)
        c = hot_redis.List(a, key=b.key)
        self.assertItemsEqual(a, c)

    def test_value(self):
        a = ["wagwaan", "hot", "skull"]
        self.assertEqual(hot_redis.List(a), a)

    def test_empty(self):
        self.assertEqual(hot_redis.List(), [])

    def test_iter(self):
        a = ["wagwaan", "hot", "skull"]
        for i, x in enumerate(hot_redis.List(a)):
            self.assertEqual(x, a[i])

    def test_add(self):
        a = ["wagwaan", "hot", "skull"]
        b = ["nba", "hang", "time"]
        self.assertEqual(a + b, hot_redis.List(a) + hot_redis.List(b))
        self.assertEqual(a + b, hot_redis.List(a) + b)
        c = hot_redis.List(a)
        d = hot_redis.List(b)
        d += c
        c += b
        self.assertEqual(a + b, c)
        self.assertEqual(b + a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_mul(self):
        a = ["wagwaan", "hot", "skull"]
        b = hot_redis.List(a)
        i = 10
        self.assertEqual(a * i, hot_redis.List(a) * i)
        b *= i
        self.assertEqual(a * i, b)

    def test_len(self):
        a = ["wagwaan", "hot", "skull"]
        self.assertEqual(len(a), len(hot_redis.List(a)))

    def test_get(self):
        a = ["wagwaan", "hot", "skull"] * 10
        b = hot_redis.List(a)
        self.assertEqual(a[4], b[4])
        self.assertEqual(a[3:12], b[3:12])
        self.assertEqual(a[:-5], b[:-5])
        self.assertRaises(IndexError, lambda: b[len(b)])

    def test_set(self):
        a = ["wagwaan", "hot", "skull"]
        b = hot_redis.List(a)
        i = "popcaan"
        a[1] = i
        self.assertNotEqual(a, b)
        b[1] = i
        self.assertEqual(a, b)
        # todo: slice

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_del(self):
        a = ["wagwaan", "hot", "skull"]
        b = hot_redis.List(a)
        del a[1]
        self.assertNotEqual(a, b)
        del b[1]
        self.assertEqual(a, b)
        # todo: slice?

    def test_contains(self):
        a = ["wagwaan", "hot", "skull"]
        b = hot_redis.List(a)
        self.assertIn("wagwaan", a)
        self.assertNotIn("hotskull", a)

    def test_extend(self):
        a = ["wagwaan", "hot", "skull"]
        b = ["nba", "hang", "time"]
        c = hot_redis.List(a)
        a.extend(b)
        c.extend(b)
        self.assertEqual(a, c)

    def test_append(self):
        a = ["wagwaan", "hot", "skull"]
        b = hot_redis.List(a)
        i = "popcaan"
        a.append(i)
        b.append(i)
        self.assertEqual(a, b)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_insert(self):
        a = ["wagwaan", "hot", "skull"]
        b = hot_redis.List(a)
        i = "popcaan"
        a.insert(1, i)
        b.insert(1, i)
        self.assertEqual(a, b)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_pop(self):
        a = ["wagwaan", "hot", "skull"] * 10
        b = hot_redis.List(a)
        a.pop()
        b.pop()
        self.assertEqual(a, b)
        a.pop(0)
        b.pop(0)
        self.assertEqual(a, b)
        a.pop(-1)
        b.pop(-1)
        self.assertEqual(a, b)
        a.pop(20)
        b.pop(20)
        self.assertEqual(a, b)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_reverse(self):
        a = ["wagwaan", "hot", "skull"]
        b = hot_redis.List(a)
        a.reverse()
        b.reverse()
        self.assertEqual(a, b)

    def test_index(self):
        a = ["wagwaan", "hot", "skull"] * 10
        b = hot_redis.List(a)
        c = "wagwaan"
        self.assertEqual(a.index(c), b.index(c))
        self.assertRaises(ValueError, lambda: b.index("popcaan"))

    def test_count(self):
        a = ["wagwaan", "hot", "skull"] * 10
        b = hot_redis.List(a)
        self.assertEqual(a.count("wagwaan"), b.count("wagwaan"))
        self.assertEqual(a.count("popcaan"), b.count("popcaan"))

    def test_sort(self):
        a = ["wagwaan", "hot", "skull"] * 10
        b = hot_redis.List(a)
        a.sort()
        b.sort()
        self.assertEqual(a, b)
        a.sort(reverse=True)
        b.sort(reverse=True)
        self.assertEqual(a, b)


class SetTests(BaseTestCase):

    def test_value(self):
        a = set(["wagwaan", "hot", "skull"])
        self.assertEqual(hot_redis.Set(a), a)

    def test_empty(self):
        self.assertEqual(hot_redis.Set(), set())

    def test_add(self):
        a = set(["wagwaan", "hot", "skull"])
        b = hot_redis.Set(a)
        i = "popcaan"
        a.add(i)
        b.add(i)
        self.assertEqual(b, a)

    def test_update(self):
        a = set(["wagwaan", "hot", "skull"])
        b = set(["nba", "hang", "time"])
        c = set(["rap", "dot", "mom"])
        d = hot_redis.Set(a)
        a.update(b, c)
        d.update(b, c)
        self.assertEqual(d, a)

    def test_pop(self):
        a = hot_redis.Set(["wagwaan", "hot", "skull"])
        i = len(a)
        b = a.pop()
        self.assertEqual(len(a), i - 1)
        self.assertNotIn(b, a)

    def test_clear(self):
        a = hot_redis.Set(["wagwaan", "hot", "skull"])
        a.clear()
        self.assertEqual(len(a), 0)

    def test_remove(self):
        a = hot_redis.Set(["wagwaan", "hot", "skull"])
        i = len(a)
        b = "wagwaan"
        a.remove(b)
        self.assertEqual(len(a), i - 1)
        self.assertNotIn(b, a)
        self.assertRaises(KeyError, lambda: a.remove("popcaan"))

    def test_discard(self):
        a = hot_redis.Set(["wagwaan", "hot", "skull"])
        i = len(a)
        b = "wagwaan"
        a.discard(b)
        self.assertEqual(len(a), i - 1)
        self.assertNotIn(b, a)
        self.assertEqual(a.discard("popcaan"), None)

    def test_len(self):
        a = set(["wagwaan", "hot", "skull"])
        b = hot_redis.Set(a)
        self.assertEqual(len(a), len(b))

    def test_contains(self):
        a = hot_redis.Set(["wagwaan", "hot", "skull"])
        self.assertIn("wagwaan", a)
        self.assertNotIn("popcaan", a)

    def test_intersection(self):
        a = set(["wagwaan", "hot", "skull"])
        b = set(["wagwaan", "flute", "don"])
        c = set(["wagwaan", "worldstar", "hiphop"])
        d = hot_redis.Set(a)
        e = a.intersection(b, c)
        self.assertEqual(a.intersection(b), d.intersection(b))
        self.assertEqual(e, d.intersection(b, c))
        self.assertEqual(e, d.intersection(hot_redis.Set(b), c))
        self.assertEqual(e, d.intersection(b, hot_redis.Set(c)))
        self.assertEqual(e, d.intersection(hot_redis.Set(b), hot_redis.Set(c)))

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_intersection_update(self):
        a = set(["wagwaan", "hot", "skull"])
        b = set(["wagwaan", "flute", "don"])
        c = set(["wagwaan", "worldstar", "hiphop"])
        d = a.copy()
        d.intersection_update(b)
        e = hot_redis.Set(a)
        e.intersection_update(b)
        self.assertEqual(e, d)
        d = a.copy()
        d.intersection_update(b, c)
        e = hot_redis.Set(a)
        e.intersection_update(b, c)
        self.assertEqual(e, d)
        e = hot_redis.Set(a)
        e.intersection_update(hot_redis.Set(b), c)
        self.assertEqual(e, d)
        e = hot_redis.Set(a)
        e.intersection_update(b, hot_redis.Set(c))
        self.assertEqual(e, d)
        e = hot_redis.Set(a)
        e.intersection_update(hot_redis.Set(b), hot_redis.Set(c))
        self.assertEqual(e, d)

    def test_difference(self):
        a = set(["wagwaan", "hot", "skull"])
        b = set(["wagwaan", "flute", "don"])
        c = set(["wagwaan", "worldstar", "hiphop"])
        d = hot_redis.Set(a)
        e = a.difference(b, c)
        self.assertEqual(a.difference(b), d.difference(b))
        self.assertEqual(e, d.difference(b, c))
        self.assertEqual(e, d.difference(hot_redis.Set(b), c))
        self.assertEqual(e, d.difference(b, hot_redis.Set(c)))
        self.assertEqual(e, d.difference(hot_redis.Set(b), hot_redis.Set(c)))

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_difference_update(self):
        a = set(["wagwaan", "hot", "skull"])
        b = set(["wagwaan", "flute", "don"])
        c = set(["wagwaan", "worldstar", "hiphop"])
        d = a.copy()
        d.difference_update(b)
        e = hot_redis.Set(a)
        e.difference_update(b)
        self.assertEqual(e, d)
        d = a.copy()
        d.difference_update(b, c)
        e = hot_redis.Set(a)
        e.difference_update(b, c)
        self.assertEqual(e, d)
        e = hot_redis.Set(a)
        e.difference_update(hot_redis.Set(b), c)
        self.assertEqual(e, d)
        e = hot_redis.Set(a)
        e.difference_update(b, hot_redis.Set(c))
        self.assertEqual(e, d)
        e = hot_redis.Set(a)
        e.difference_update(hot_redis.Set(b), hot_redis.Set(c))
        self.assertEqual(e, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_symmetric_difference(self):
        a = set(["wagwaan", "hot", "skull"])
        b = set(["wagwaan", "flute", "don"])
        c = hot_redis.Set(a)
        d = a.symmetric_difference(b)
        self.assertEqual(d, c.symmetric_difference(b))
        self.assertEqual(d, c.symmetric_difference(hot_redis.Set(b)))
        self.assertEqual(d, a.symmetric_difference(hot_redis.Set(b)))

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_symmetric_difference_update(self):
        a = set(["wagwaan", "hot", "skull"])
        b = set(["wagwaan", "flute", "don"])
        c = a.copy()
        c.difference_update(b)
        d = hot_redis.Set(a)
        d.difference_update(b)
        self.assertEqual(d, c)
        d = hot_redis.Set(a)
        d.difference_update(hot_redis.Set(b))
        self.assertEqual(d, c)

    def test_disjoint(self):
        a = set(["wagwaan", "hot", "skull"])
        b = hot_redis.Set(a)
        c = hot_redis.Set(["wagwaan", "flute", "don"])
        d = set(["nba", "hang", "time"])
        e = hot_redis.Set(d)
        self.assertFalse(b.isdisjoint(a))
        self.assertFalse(b.isdisjoint(c))
        self.assertTrue(b.isdisjoint(d))
        self.assertTrue(b.isdisjoint(e))

    def test_cmp(self):
        a = set(["wagwaan", "hot", "skull"])
        b = set(["nba", "hang", "time"])
        c = hot_redis.Set(a)
        d = hot_redis.Set(b)
        self.assertEqual(a > b, c > d)
        self.assertEqual(a < b, c < d)
        self.assertEqual(a > b, c > b)
        self.assertEqual(a < b, c < b)
        self.assertEqual(a >= b, c >= d)
        self.assertEqual(a <= b, c <= d)
        self.assertEqual(a >= b, c >= b)
        self.assertEqual(a <= b, c <= b)
        self.assertEqual(a.issubset(b), c.issubset(d))
        self.assertEqual(a.issuperset(b), c.issuperset(d))
        self.assertEqual(a.issubset(b), c.issubset(b))
        self.assertEqual(a.issuperset(b), c.issuperset(b))


class DictTests(BaseTestCase):

    def test_value(self):
        a = {"wagwaan": "popcaan", "flute": "don"}
        self.assertEqual(hot_redis.Dict(a), a)

    def test_empty(self):
        self.assertEqual(hot_redis.Dict(), {})

    def test_update(self):
        a = {"wagwaan": "popcaan", "flute": "don"}
        b = {"wagwaan": "hotskull", "nba": "hangtime"}
        c = hot_redis.Dict(a)
        a.update(b)
        c.update(b)
        self.assertEqual(a, c)

    def test_iter(self):
        a = {"wagwaan": "popcaan", "flute": "don"}
        self.assertItemsEqual(iter(a), iter(hot_redis.Dict(a)))

    def test_keys(self):
        a = {"wagwaan": "popcaan", "flute": "don"}
        self.assertItemsEqual(a.keys(), hot_redis.Dict(a).keys())

    def test_values(self):
        a = {"wagwaan": "popcaan", "flute": "don"}
        self.assertItemsEqual(a.values(), hot_redis.Dict(a).values())

    def test_items(self):
        a = {"wagwaan": "popcaan", "flute": "don"}
        self.assertItemsEqual(a.items(), hot_redis.Dict(a).items())

    def test_setdefault(self):
        a = {"wagwaan": "popcaan", "flute": "don"}
        b = hot_redis.Dict(a)
        c = "nba"
        d = "hangtime"
        e = b.setdefault(c, d)
        self.assertEqual(e, d)
        self.assertEqual(b[c], d)
        self.assertEqual(a.setdefault(c, d), e)
        e = b.setdefault(c, c)
        self.assertEqual(e, d)
        self.assertEqual(a.setdefault(c, c), e)

    def test_get(self):
        a = {"wagwaan": "popcaan", "flute": "don"}
        b = hot_redis.Dict(a)
        self.assertEqual(a["wagwaan"], b["wagwaan"])
        self.assertEqual(a.get("wagwaan"), b.get("wagwaan"))
        self.assertRaises(KeyError, lambda: b["hotskull"])
        self.assertEqual(a.get("hotskull"), b.get("hotskull"))
        self.assertEqual(a.get("hotskull", "don"), b.get("hotskull", "don"))
        self.assertNotEqual(a.get("hotskull", "don"), b.get("hotskull", "x"))

    def test_set(self):
        a = hot_redis.Dict({"wagwaan": "popcaan", "flute": "don"})
        a["wagwaan"] = "hotskull"
        self.assertEqual(a["wagwaan"], "hotskull")

    def test_del(self):
        a = hot_redis.Dict({"wagwaan": "popcaan", "flute": "don"})
        del a["wagwaan"]
        self.assertRaises(KeyError, lambda: a["wagwaan"])
        def del_missing():
            del a["hotskull"]
        self.assertRaises(KeyError, del_missing)

    def test_len(self):
        a = {"wagwaan": "popcaan", "flute": "don"}
        b = hot_redis.Dict(a)
        self.assertEqual(len(a), len(b))

    def test_contains(self):
        a = {"wagwaan": "popcaan", "flute": "don"}
        b = hot_redis.Dict(a)
        self.assertIn("wagwaan", a)
        self.assertNotIn("hotskull", a)

    def test_copy(self):
        a = hot_redis.Dict({"wagwaan": "popcaan", "flute": "don"})
        b = a.copy()
        self.assertEqual(type(a), type(b))
        self.assertNotEqual(a.key, b.key)

    def test_clear(self):
        a = hot_redis.Dict({"wagwaan": "popcaan", "flute": "don"})
        a.clear()
        self.assertEqual(len(a), 0)

    def test_fromkeys(self):
        a = ["wagwaan", "hot", "skull"]
        b = "popcaan"
        c = hot_redis.Dict.fromkeys(a)
        self.assertItemsEqual(sorted(a), sorted(c.keys()))
        self.assertFalse(c["wagwaan"])
        c = hot_redis.Dict.fromkeys(a, b)
        self.assertEqual(c["wagwaan"], b)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_defaultdict(self):
        a = "wagwaan"
        b = "popcaan"
        c = hot_redis.DefaultDict(lambda: b)
        self.assertEqual(c[a], b)
        c[b] += a
        self.assertEqual(c[b], b + a)


class StringTests(BaseTestCase):

    def test_value(self):
        a = "wagwaan"
        self.assertEqual(hot_redis.String(a), a)

    def test_empty(self):
        self.assertEqual(hot_redis.String(), "")

    def test_add(self):
        a = "wagwaan"
        b = "hotskull"
        c = hot_redis.String(a)
        d = hot_redis.String(b)
        self.assertEqual(a + b, hot_redis.String(a) + hot_redis.String(b))
        self.assertEqual(a + b, hot_redis.String(a) + b)
        d += c
        c += b
        self.assertEqual(a + b, c)
        self.assertEqual(b + a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_mul(self):
        a = "wagwaan"
        b = hot_redis.String(a)
        i = 9000
        self.assertEqual(a * i, hot_redis.String(a) * i)
        b *= i
        self.assertEqual(a * i, b)

    def test_len(self):
        a = "wagwaan"
        self.assertEqual(len(a), len(hot_redis.String(a)))

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_set(self):
        a = "wagwaan hotskull"
        b = "flute don"
        for i in range(0, len(b)):
            for j in range(i, len(b)):
                c = list(a)
                d = hot_redis.String(a)
                c[i:j] = list(b)
                d[i:j] = b
                c = "".join(c)
                self.assertEqual(d, c)

    def test_get(self):
        a = "wagwaan hotskull"
        b = hot_redis.String(a)
        self.assertEqual(a[4], b[4])
        self.assertEqual(a[3:12], b[3:12])
        self.assertEqual(a[:-5], b[:-5])
        self.assertRaises(IndexError, lambda: b[len(b)])

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_mutability(self):
        a = "wagwaan hotskull"
        b = "flute don"
        c = hot_redis.String(a)
        d = hot_redis.ImmutableString(a)
        keyC = c.key
        keyD = d.key
        a += b
        c += b
        d += b
        self.assertEqual(a, c)
        self.assertEqual(a, d)
        self.assertEqual(c.key, keyC)
        self.assertNotEqual(d.key, keyD)
        keyD = d.key
        i = 9000
        a *= i
        c *= i
        d *= i
        self.assertEqual(a, c)
        self.assertEqual(a, d)
        self.assertEqual(c.key, keyC)
        self.assertNotEqual(d.key, keyD)
        def immutable_set():
            d[0] = b
        self.assertRaises(TypeError, immutable_set)


class IntTests(BaseTestCase):

    def test_value(self):
        a = 420
        self.assertEqual(hot_redis.Int(a), a)

    def test_empty(self):
        self.assertEqual(hot_redis.Int(), 0)

    def test_set(self):
        a = 420
        b = hot_redis.Int()
        b.value = a
        self.assertEqual(b, a)
        b.value = 0
        self.assertEqual(b, 0)

    def test_add(self):
        a = 420
        b = 9000
        self.assertEqual(a + b, hot_redis.Int(a) + hot_redis.Int(b))
        self.assertEqual(a + b, hot_redis.Int(a) + b)
        c = hot_redis.Int(a)
        d = hot_redis.Int(b)
        d += c
        c += b
        self.assertEqual(a + b, c)
        self.assertEqual(b + a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_mul(self):
        a = 420
        b = hot_redis.Int(a)
        i = 9000
        self.assertEqual(a * i, hot_redis.Int(a) * i)
        b *= i
        self.assertEqual(a * i, b)

    def test_sub(self):
        a = 420
        b = 9000
        self.assertEqual(a - b, hot_redis.Int(a) - hot_redis.Int(b))
        self.assertEqual(a - b, hot_redis.Int(a) - b)
        c = hot_redis.Int(a)
        d = hot_redis.Int(b)
        d -= c
        c -= b
        self.assertEqual(a - b, c)
        self.assertEqual(b - a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_div(self):
        a = 420
        b = 9000
        self.assertEqual(a / b, hot_redis.Int(a) / hot_redis.Int(b))
        self.assertEqual(a / b, hot_redis.Int(a) / b)
        c = hot_redis.Int(a)
        d = hot_redis.Int(b)
        d /= c
        c /= b
        self.assertEqual(a / b, c)
        self.assertEqual(b / a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_mod(self):
        a = 420
        b = 9000
        self.assertEqual(a % b, hot_redis.Int(a) % hot_redis.Int(b))
        self.assertEqual(a % b, hot_redis.Int(a) % b)
        c = hot_redis.Int(a)
        d = hot_redis.Int(b)
        d %= c
        c %= b
        self.assertEqual(a % b, c)
        self.assertEqual(b % a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_pow(self):
        a = 4
        b = 20
        self.assertEqual(a ** b, hot_redis.Int(a) ** hot_redis.Int(b))
        self.assertEqual(a ** b, hot_redis.Int(a) ** b)
        c = hot_redis.Int(a)
        d = hot_redis.Int(b)
        d **= c
        c **= b
        self.assertEqual(a ** b, c)
        self.assertEqual(b ** a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_and(self):
        a = 420
        b = 9000
        self.assertEqual(a & b, hot_redis.Int(a) & hot_redis.Int(b))
        self.assertEqual(a & b, hot_redis.Int(a) & b)
        c = hot_redis.Int(a)
        d = hot_redis.Int(b)
        d &= c
        c &= b
        self.assertEqual(a & b, c)
        self.assertEqual(b & a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_or(self):
        a = 420
        b = 9000
        self.assertEqual(a | b, hot_redis.Int(a) | hot_redis.Int(b))
        self.assertEqual(a | b, hot_redis.Int(a) | b)
        c = hot_redis.Int(a)
        d = hot_redis.Int(b)
        d |= c
        c |= b
        self.assertEqual(a | b, c)
        self.assertEqual(b | a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_xor(self):
        a = 420
        b = 9000
        self.assertEqual(a ^ b, hot_redis.Int(a) ^ hot_redis.Int(b))
        self.assertEqual(a ^ b, hot_redis.Int(a) ^ b)
        c = hot_redis.Int(a)
        d = hot_redis.Int(b)
        d ^= c
        c ^= b
        self.assertEqual(a ^ b, c)
        self.assertEqual(b ^ a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_lshift(self):
        a = 4
        b = 20
        self.assertEqual(a << b, hot_redis.Int(a) << hot_redis.Int(b))
        self.assertEqual(a << b, hot_redis.Int(a) << b)
        c = hot_redis.Int(a)
        d = hot_redis.Int(b)
        d <<= c
        c <<= b
        self.assertEqual(a << b, c)
        self.assertEqual(b << a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_rshift(self):
        a = 9000
        b = 4
        self.assertEqual(a >> b, hot_redis.Int(a) >> hot_redis.Int(b))
        self.assertEqual(a >> b, hot_redis.Int(a) >> b)
        c = hot_redis.Int(a)
        d = hot_redis.Int(b)
        d >>= c
        c >>= b
        self.assertEqual(a >> b, c)
        self.assertEqual(b >> a, d)


class FloatTests(BaseTestCase):

    def test_value(self):
        a = 420.666
        self.assertAlmostEqual(hot_redis.Float(a), a)

    def test_empty(self):
        self.assertEqual(hot_redis.Int(), .0)

    def test_set(self):
        a = 420.666
        b = hot_redis.Float()
        b.value = a
        self.assertAlmostEqual(b, a)
        b.value = .0
        self.assertAlmostEqual(b, .0)

    def test_add(self):
        a = 420.666
        b = 9000.666
        self.assertAlmostEqual(a + b, hot_redis.Float(a) + hot_redis.Float(b))
        self.assertAlmostEqual(a + b, hot_redis.Float(a) + b)
        c = hot_redis.Float(a)
        d = hot_redis.Float(b)
        d += c
        c += b
        self.assertAlmostEqual(a + b, c)
        self.assertAlmostEqual(b + a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_mul(self):
        a = 420.666
        b = hot_redis.Float(a)
        i = 9000.666
        self.assertAlmostEqual(a * i, hot_redis.Float(a) * i)
        b *= i
        self.assertAlmostEqual(a * i, b)

    def test_sub(self):
        a = 420.666
        b = 9000.666
        self.assertAlmostEqual(a - b, hot_redis.Float(a) - hot_redis.Float(b))
        self.assertAlmostEqual(a - b, hot_redis.Float(a) - b)
        c = hot_redis.Float(a)
        d = hot_redis.Float(b)
        d -= c
        c -= b
        self.assertAlmostEqual(a - b, c)
        self.assertAlmostEqual(b - a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_div(self):
        a = 420.666
        b = 9000.666
        self.assertAlmostEqual(a / b, hot_redis.Float(a) / hot_redis.Float(b))
        self.assertAlmostEqual(a / b, hot_redis.Float(a) / b)
        c = hot_redis.Float(a)
        d = hot_redis.Float(b)
        d /= c
        c /= b
        self.assertAlmostEqual(a / b, c)
        self.assertAlmostEqual(b / a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_mod(self):
        a = 420.666
        b = 9000.666
        self.assertAlmostEqual(a % b, hot_redis.Float(a) % hot_redis.Float(b))
        self.assertAlmostEqual(a % b, hot_redis.Float(a) % b)
        c = hot_redis.Float(a)
        d = hot_redis.Float(b)
        d %= c
        c %= b
        self.assertAlmostEqual(a % b, c)
        self.assertAlmostEqual(b % a, d)

    @unittest.skipIf(TEST_NO_LUA, "No Lua")
    def test_pow(self):
        a = 4.666
        b = 20.666
        c = 4
        d = 2
        self.assertAlmostEqual(a ** b, hot_redis.Float(a) ** hot_redis.Float(b))
        self.assertAlmostEqual(a ** b, hot_redis.Float(a) ** b)
        e = hot_redis.Float(a)
        f = hot_redis.Float(b)
        f **= c
        e **= d
        self.assertAlmostEqual(a ** d, e)
        self.assertAlmostEqual(b ** c, f)


@unittest.skipIf(TEST_NO_LUA, "No Lua")
class QueueTests(BaseTestCase):

    def test_put(self):
        a = "wagwaan"
        b = "hotskull"
        q = hot_redis.Queue(maxsize=2)
        q.put(a)
        self.assertIn(a, q)
        q.put(b)
        self.assertIn(b, q)
        self.assertRaises(hot_redis.queue.Full, lambda: q.put("popcaan", block=False))
        start = time.time()
        timeout = 2
        try:
            q.put("popcaan", timeout=timeout)
        except hot_redis.queue.Full:
            pass
        self.assertTrue(time.time() - start >= timeout)

    def test_get(self):
        a = "wagwaan"
        b = "hotskull"
        q = hot_redis.Queue()
        q.put(a)
        q.put(b)
        self.assertEqual(a, q.get())
        self.assertNotIn(a, q)
        self.assertEqual(b, q.get())
        self.assertNotIn(b, q)
        self.assertRaises(hot_redis.queue.Empty, lambda: q.get(block=False))
        start = time.time()
        timeout = 2
        try:
            q.get(timeout=timeout)
        except hot_redis.queue.Empty:
            pass
        self.assertTrue(time.time() - start >= timeout)

    def test_empty(self):
        q = hot_redis.Queue()
        self.assertTrue(q.empty())
        q.put("wagwaan")
        self.assertFalse(q.empty())
        q.get()
        self.assertTrue(q.empty())

    def test_full(self):
        q = hot_redis.Queue(maxsize=2)
        self.assertFalse(q.full())
        q.put("wagwaan")
        self.assertFalse(q.full())
        q.put("hotskull")
        self.assertTrue(q.full())
        q.get()
        self.assertFalse(q.full())

    def test_size(self):
        q = hot_redis.Queue()
        self.assertEqual(q.qsize(), 0)
        q.put("wagwaan")
        self.assertEqual(q.qsize(), 1)
        q.put("hotskull")
        self.assertEqual(q.qsize(), 2)
        q.get()
        self.assertEqual(q.qsize(), 1)

    def test_lifo_queue(self):
        a = "wagwaan"
        b = "hotskull"
        q = hot_redis.LifoQueue()
        q.put(a)
        q.put(b)
        self.assertEqual(b, q.get())
        self.assertNotIn(b, q)
        self.assertEqual(a, q.get())
        self.assertNotIn(a, q)

    def test_set_queue(self):
        a = "wagwaan"
        q = hot_redis.SetQueue()
        q.put(a)
        self.assertEqual(q.qsize(), 1)
        q.put(a)
        self.assertEqual(q.qsize(), 1)
        self.assertEqual(q.get(), a)
        self.assertEqual(q.qsize(), 0)


@unittest.skipIf(TEST_NO_LUA, "No Lua")
class CounterTests(BaseTestCase):

    def test_value(self):
        a = "wagwaan"
        b = {"hot": 420, "skull": -9000}
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        e = collections.Counter(**b)
        f = hot_redis.MultiSet(**b)
        self.assertEqual(d, c)
        self.assertEqual(f, e)

    def test_empty(self):
        self.assertEqual(hot_redis.MultiSet(), collections.Counter())

    def test_values(self):
        a = "wagwaan"
        b = {"hot": 420, "skull": -9000}
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        e = collections.Counter(**b)
        f = hot_redis.MultiSet(**b)
        self.assertItemsEqual(c.values(), d.values())
        self.assertItemsEqual(e.values(), f.values())

    def test_get(self):
        a = "wagwaan"
        b = {"hot": 420, "skull": -9000}
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        e = collections.Counter(**b)
        f = hot_redis.MultiSet(**b)
        self.assertEqual(c.get("a"), d.get("a"))
        self.assertEqual(c.get("flute", "don"), d.get("flute", "don"))
        self.assertEqual(e.get("hot"), f.get("hot"))
        self.assertEqual(e.get("skull"), f.get("skull"))
        self.assertEqual(e.get("flute", "don"), e.get("flute", "don"))

    def test_del(self):
        a = hot_redis.MultiSet("wagwaan")
        del a["hotskull"]

    def test_update(self):
        a = "wagwaan"
        b = {"hotskull": 420}
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c.update(hot_redis.MultiSet(a))
        d.update(hot_redis.MultiSet(a))
        self.assertEqual(d, c)
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c.update(collections.Counter(a))
        d.update(collections.Counter(a))
        self.assertEqual(d, c)
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c.update(a)
        d.update(a)
        self.assertEqual(d, c)
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c.update(b)
        d.update(b)
        self.assertEqual(d, c)
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c.update(**b)
        d.update(**b)
        self.assertEqual(d, c)

    def test_subtract(self):
        a = "wagwaan"
        b = {"hotskull": 420}
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c.subtract(hot_redis.MultiSet(a))
        d.subtract(hot_redis.MultiSet(a))
        self.assertEqual(d, c)
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c.subtract(collections.Counter(a))
        d.subtract(collections.Counter(a))
        self.assertEqual(d, c)
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c.subtract(a)
        d.subtract(a)
        self.assertEqual(d, c)
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c.subtract(b)
        d.subtract(b)
        self.assertEqual(d, c)
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c.subtract(**b)
        d.subtract(**b)
        self.assertEqual(d, c)

    def test_intersection(self):
        a = "wagwaan"
        b = "flute don"
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c &= hot_redis.MultiSet(b)
        d &= hot_redis.MultiSet(b)
        self.assertEqual(d, c)
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c &= collections.Counter(b)
        d &= collections.Counter(b)
        self.assertEqual(d, c)

    def test_union(self):
        a = "wagwaan"
        b = "flute don"
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c |= hot_redis.MultiSet(b)
        d |= hot_redis.MultiSet(b)
        self.assertEqual(d, c)
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        c |= collections.Counter(b)
        d |= collections.Counter(b)
        self.assertEqual(d, c)

    def test_elements(self):
        a = "wagwaan"
        b = {"hotskull": 420}
        c = collections.Counter(a)
        d = hot_redis.MultiSet(a)
        e = collections.Counter(**b)
        f = hot_redis.MultiSet(**b)
        self.assertItemsEqual(sorted(c.elements()), sorted(d.elements()))
        self.assertItemsEqual(sorted(e.elements()), sorted(f.elements()))

    def test_most_common(self):
        a = "wagwaan"
        b = collections.Counter(a)
        c = hot_redis.MultiSet(a)
        d = 420
        check = b.most_common(d)
        for i, e in enumerate(c.most_common(d)):
            self.assertEqual(e[1], check[i][1])
        check = b.most_common()
        for i, e in enumerate(c.most_common()):
            self.assertEqual(e[1], check[i][1])

@unittest.skipIf(TEST_NO_LUA, "No Lua")
class TransactionTests(BaseTestCase):

    def test_transaction(self):
        with_transaction = hot_redis.List([1])
        without_transaction = hot_redis.List(key=with_transaction.key,
                                             client=hot_redis.HotClient())
        with hot_redis.transaction():
            with_transaction.append(1)
            self.assertEqual(len(without_transaction), 1)
        self.assertEqual(len(without_transaction), 2)


@unittest.skipIf(TEST_NO_LUA, "No Lua")
class LockTests(BaseTestCase):

    def test_semaphore(self):
        semaphore = hot_redis.Semaphore()
        self.assertEqual(semaphore.acquire(), True)
        self.assertEqual(semaphore.release(), None)
        self.assertEqual(semaphore.release(), None)

    def test_bounded_semaphore(self):
        max_size = 2
        semaphore = hot_redis.BoundedSemaphore(value=max_size)
        self.assertEqual(semaphore.acquire(), True)
        self.assertEqual(semaphore.release(), None)
        with semaphore:
            with semaphore:
                self.assertEqual(semaphore.acquire(block=False), False)
        self.assertRaises(RuntimeError, semaphore.release)

    def test_lock(self):
        lock = hot_redis.Lock()
        self.assertEqual(lock.acquire(), True)
        self.assertEqual(lock.release(), None)
        with lock:
            self.assertEqual(lock.acquire(block=False), False)
        self.assertRaises(RuntimeError, lock.release)


if __name__ == "__main__":
    unittest.main()
